@@ -4,6 +4,8 @@ use warnings;
 use strict;
 #use parent 'Keyring';
 use Net::DBus;
+use Try::Tiny;
+use Proc::SyncExec qw/sync_exec/;
 use Carp;
 
 =head1 NAME
@@ -12,15 +14,20 @@ Passwd::Keyring::KDEWallet - Password storage implementation based on KDE Wallet
 
 =head1 VERSION
 
-Version 0.2008
+Version 0.60
 
 =cut
 
-our $VERSION = '0.2008';
+our $VERSION = '0.60';
 
 our $APP_NAME = "Passwd::Keyring";
 our $FOLDER_NAME = "Perl-Passwd-Keyring";
 
+# Max time we wait for kwalletd
+our $KWALLETD_START_TIMEOUT = 2.0;
+# Frequency of checks
+our $KWALLETD_CHECK_FREQUENCY = 0.1;
+
 =head1 SYNOPSIS
 
 KDE Wallet based implementation of L<Passwd::Keyring>.
@@ -53,17 +60,42 @@ C<Passwd::Keyring::Auto> package).
 
 =head1 SUBROUTINES/METHODS
 
-=head2 new(app=>'app name', group=>'passwords folder')
+=head2 new
 
-Initializes the processing. Croaks if kwallet (or d-bus, or anything needed) does not 
-seem to be available.
+    Passwd::Keyring::KDEWallet->new(
+        app=>'app name', group=>'passwords folder');
 
-Handled named parameters: 
+    Passwd::Keyring::KDEWallet->new(
+        app=>'app name', group=>'passwords folder',
+        start_kwalletd_if_missing=>1);
 
-- app - symbolic application name (used in "Application .... is asking to open the wallet"
-  KDE Wallet prompt)
+Initializes the processing. Croaks if kwallet (or d-bus, or anything
+needed) does not seem to be available.
 
-- group - name for the password group (used as KDE Wallet folder name)
+Handled named parameters:
+
+=over 4
+
+=item app
+
+symbolic application name (used in "Application .... is asking
+to open the wallet" KDE Wallet prompt)
+
+=item group
+
+name for the password group (used as KDE Wallet folder name)
+
+=item dont_start_daemon
+
+by default, in case kwalletd service is missing, we try to start
+it, this option disables this behaviour
+
+=item kwalletd_path
+
+path to kwalletd binary, used in case we try starting it. Default:
+C<kwalletd> (relative path means searching in C<PATH>).
+
+=back
 
 =cut
 
@@ -73,35 +105,88 @@ sub new {
     my $self = {};
     $self->{app} = $args{app} || 'Passwd::Keyring::KDEWallet';
     $self->{group} = $args{group} || 'Passwd::Keyring::default';
-    bless $self;
+    $self->{dont_start_daemon} = $args{dont_start_daemon} || '';
+    $self->{kwalletd_path} = $args{kwalletd_path} || 'kwalletd';
+    bless $self, $cls;
 
     #$self->{bus} = Net::DBus->find()
     $self->{bus} = Net::DBus->session()
       or croak("KWallet not available (can't access DBus)");
-    # get_service also may fail by itself, I got cpantesters reports with message
-    # "org.freedesktop.DBus.Error.ServiceUnknown: The name org.kde.kwalletd was not provided by any .service files"
-    # Let's rewrite this slightly
-    my$kwallet_svc;
-    eval {
-        $kwallet_svc = $self->{bus}->get_service('org.kde.kwalletd');
-    };
-    if($@) {
-        croak("KWallet not available (not installed?). Details:\n$@");
-    } elsif (! $kwallet_svc) {
-        croak("KWallet not available (can't access KWallet, likely kwalletd not running)");
-    }
-    $self->{kwallet} = $kwallet_svc->get_object('/modules/kwalletd', 'org.kde.KWallet')
-      or croak("Kwallet not available (can't find wallet)");
+
+    $self->_init_kwallet();
+
     $self->_open_if_not_open();
 
     unless($self->{kwallet}->hasFolder($self->{handle}, $self->{group}, $self->{app})) {
         $self->{kwallet}->createFolder($self->{handle}, $self->{group}, $self->{app})
-          or croak("Failed to create $self->{group} folder (app $self->{app})");
+          or croak("Failed to create or access $self->{group} folder (app $self->{app}).\nDid you reject prompt to open the wallet?\n");
     }
 
     return $self;
 }
 
+# Called from the constructor, setups self->{kwallet} attribute (top level
+# service object)
+sub _init_kwallet {
+    my $self = shift;
+
+    # get_service may fail by itself, if kwalletd is down, in some
+    # cases it fails with
+    #
+    # org.freedesktop.DBus.Error.ServiceUnknown: The name org.kde.kwalletd was not provided by any .service files
+
+    my $kwallet_svc;
+    my $error;
+    try {
+        $kwallet_svc = $self->{bus}->get_service('org.kde.kwalletd');
+    } catch {
+        $error = $_;
+        chomp($error);
+    };
+
+    unless($kwallet_svc) {
+        # Mayhaps we are allowed to start kwalletd?
+        if(! $self->{dont_start_daemon}
+             && $error =~ /^org\.freedesktop\.DBus\.Error\.ServiceUnknown:/) {
+            print STDERR "KWallet service not available, attempting to start $self->{kwalletd_path}\n";
+            # spawn kwalletd
+            my $pid = sync_exec
+                sub {
+                    # Without this prove (tests) hang
+                    close (STDOUT);
+                    return 1;
+                },
+                $self->{kwalletd_path};
+            unless($pid) {
+                croak "KWallet not available (not installed or not started)\nand attempt to start it failed\nOriginal error:\n$error" . "Attempted command:\n" . $self->{kwalletd_path} . "\nCommand failure: $!\n";
+            };
+            my $error2;
+            my $DEADLINE = time() + $KWALLETD_START_TIMEOUT;
+            # while($proc->alive && time() <= $DEADLINE) {
+            while(time() <= $DEADLINE) {
+                try {
+                    $kwallet_svc = $self->{bus}->get_service('org.kde.kwalletd');
+                } catch {
+                    $error2 = $_;
+                    chomp($error2);
+                    # warn "Still not available: $@\n";
+                };
+                last if $kwallet_svc;
+                sleep($KWALLETD_CHECK_FREQUENCY);
+            }
+            unless($kwallet_svc) {
+                croak "KWallet not available (not installed or not started),\nand attempt to start it did not help\nFirst error:\n$error\nLast error:\n$error2\nAttempted command: $self->{kwalletd_path}\n";
+            }
+        } else {
+            croak "KWallet not available (not installed or not started),\nand we are forbidden to start it ourselves.\nOriginal error:\n$error\n";
+        }
+    }
+
+    $self->{kwallet} = $kwallet_svc->get_object(
+        '/modules/kwalletd', 'org.kde.KWallet')
+      or croak("Kwallet not available (can't find wallet)");
+}
+
 sub _open_if_not_open {
     my $self = shift;
 
@@ -118,7 +203,7 @@ sub _open_if_not_open {
 
 =head2 set_password(username, password, realm)
 
-Sets (stores) password identified by given realm for given user 
+Sets (stores) password identified by given realm for given user
 
 =cut
 
@@ -191,7 +276,7 @@ Approach inspired by L<http://www.perlmonks.org/?node_id=869620>.
 
 =head1 BUGS
 
-Please report any bugs or feature requests to 
+Please report any bugs or feature requests to
 issue tracker at L<https://bitbucket.org/Mekk/perl-keyring-kdewallet>.
 
 =head1 SUPPORT
@@ -222,5 +307,3 @@ See http://dev.perl.org/licenses/ for more information.
 
 
 1; # End of Passwd::Keyring::KDEWallet
-
-
